' This program is free software; you can redistribute it and/or
' modify it under the terms of the GNU General Public License
' as published by the Free Software Foundation; either version 2
' of the License, or (at your option) any later version.

' This program is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
' GNU General Public License for more details.

' You should have received a copy of the GNU General Public License
' along with this program (it is printed by the DisplayLicense()
' subroutine, and included in a file called license.txt with all
' standard distributions); if not, write to the Free Software
' Foundation, Inc., 59 Temple Place - Suite 330, Boston,
' MA  02111-1307, USA.

' vCardSplit is copyright (C) Philip Storry, 2004.


#COMPILE EXE         ' Compile to an EXE
'#DEBUG ERROR ON      ' Include debugging code when compiling

#RESOURCE "resources.pbr" ' Include the icon etc...

#DIM ALL             ' Forces all variables to be declared

' Declare subs and functions used outside PBMAIN

DECLARE FUNCTION FormatFileName (unformatted AS STRING) AS STRING

DECLARE SUB DisplayHelp ()
DECLARE SUB DisplayCopyright ()
DECLARE SUB DisplayLicense ()

' Macros that are useful in this program

MACRO utrim (p1) = UCASE$(TRIM$(p1))
MACRO stripNum (p1) = LTRIM$(STR$(p1))
MACRO NoFile(fname) = (LEN(DIR$(fname)) = 0)

' A set of equates - mostly for build management purposes

%MajorVersion = 1                    ' Major program version
%MinorVersion = 0                    ' Minor program version
%RevisionNumber = 2                  ' Revision number (bugfixes)
%BuildNumber = 5                     ' Build number (used for testing)
$ProgramName = "vCardSplit"          ' Program name
$ProgramAuthor = "Philip Storry"     ' Program author(s)
$ProgramDate = "2004"                ' Date for copyright
%DevelopmentState = 3                ' State - 1,2,3;Alpha,Beta,Gold

FUNCTION PBMAIN

    ' Declare our local variables

    LOCAL cmd             AS STRING
    LOCAL fileHnd         AS LONG
    LOCAL outputFileHnd   AS LONG
    LOCAL lastErr         AS LONG
    LOCAL bufferIn        AS STRING
    LOCAL bufferOut       AS STRING
    LOCAL tmpStr          AS STRING
    LOCAL position        AS QUAD
    LOCAL vCardName       AS STRING
    LOCAL vCardOrg        AS STRING
    LOCAL vCardFileOut    AS STRING
    LOCAL skipvCard       AS LONG
    LOCAL currentLine     AS LONG
    LOCAL vCardEOF        AS LONG

    LOCAL cardsFound      AS LONG
    LOCAL cardsSkipped    AS LONG
    LOCAL cardsWritten    AS LONG
    LOCAL cardsUnfinished AS LONG
    LOCAL cardsUnusable   AS LONG
    LOCAL cardsExist      AS LONG
    LOCAL cardsWriteError AS LONG

    LOCAL readCycleReset  AS LONG

    ' Start!

    ' Check the command line - print help and quit if it's empty

    cmd = TRIM$(REMOVE$(COMMAND$, CHR$(34)))  ' Remove quotes...

    IF cmd = "" THEN
        CALL DisplayHelp ()
        EXIT FUNCTION
    ELSEIF LCASE$(cmd) = "license" THEN
        CALL DisplayLicense ()
        EXIT FUNCTION
    END IF
    
    ' There's something on the command line - display the copyright, and continue...
    
    CALL DisplayCopyright ()

    ON ERROR RESUME NEXT              ' If an error occurs, we must trap it
                                      ' ourselves

    fileHnd = FREEFILE                ' Allocate a free file handle

    ERRCLEAR                          ' Clear the error code (if any)

    STDOUT "  Opening " & CHR$(34) & cmd & CHR$(34);

    OPEN cmd FOR INPUT _              ' Open the file for reading...
             ACCESS READ _            ' Ensure we can only read to prevent accidents
             LOCK WRITE _             ' Lock the file to prevent other processes writing to it
             AS fileHnd

    lastErr = ERR
    IF lastErr > 0 THEN
        STDOUT ""
        STDOUT "An error occurred whilst trying to open the file!"
        EXIT FUNCTION
    END IF

    STDOUT " - opened successfully"

    ' First things first - is this actually a vCard file or not?
    ' Let's look for BEGIN:VCARD.
    ' Skip any blank lines until something is found.
    ' If we don't find it, then it's quitting time!
    
    ' Feedback is good... Note the trailing semicolon -
    ' status will be returned on the same line!
    STDOUT "  Checking file validity...";
    
    ' Clear the built in error report variable
    ERRCLEAR
    
    ' Read a line of code
    LINE INPUT# fileHnd, bufferIn
    
    ' Did we encounter an error?
    lastErr = ERR
    IF lastErr > 0 THEN
        STDOUT "  Error!"
        STDOUT "  Unable to read the first line of the file."
        STDOUT "  Perhaps the file isn't a text file, or has no content."
        CLOSE fileHnd
        EXIT FUNCTION
    END IF
    
    ' If the line was blank, then we need to skip it until we find a non-blank line.
    DO UNTIL utrim$(bufferIn) <> ""
        LINE INPUT# fileHnd, bufferIn
        ' If we've gone past the end of the file, then quit!
        IF ERR = 62 THEN
            STDOUT "  End of file encountered!"
            STDOUT "   The end of the file was encountered before a vCard was found!"
            STDOUT "   Are you sure that this is a vCard file?"
            EXIT FUNCTION
        END IF     
    LOOP
    
    
    'Is this line the start of a vCard?
        
    IF utrim$(bufferIn) = "BEGIN:VCARD" THEN
        STDOUT " File appears to contain at least one vCard"
        STDOUT "  Proceeding with splitting operation"
    ELSE
        ' It's not a vCard - but is it a UNIX text file?
        IF INSTR(UCASE$(bufferIn), "BEGIN:VCARD" & $LF) THEN
            STDOUT "  No CR/LF pair was found!"
            #IF %DevelopmentState < 3
                STDOUT "   ##DEBUG: Line length was " & stripNum(LEN(bufferIn))
                STDOUT "   ##DEBUG: INSTR function returned " & stripNum(INSTR(0, UCASE$(bufferIn), "BEGIN:VCARD" & $LF))
            #ENDIF
            STDOUT "   As a Windows program, vCardSplit expects text files with CR/LF"
            STDOUT "   line endings - UNIX files use only a LF."
            STDOUT "   If this file came from a UNIX (or modern Apple Macintosh) system,"
            STDOUT "   then please check that it was saved in a DOS or Windows text file"
            STDOUT "   format."
        ELSE
            STDOUT "  Error!"
            #IF %DevelopmentState < 3
                STDOUT "   ##DEBUG: Line length was " & stripNum(LEN(bufferIn))
            #ENDIF
            STDOUT "   Unable to read the first line of the file."
            STDOUT "   Perhaps the file isn't a text file, or has no content."
        END IF
        CLOSE fileHnd
        EXIT FUNCTION
    END IF

    ' Having found that this file contains a vCard, we now reset the read position
    ' This just makes the logic within the loop a little easier.
    ' It is not due to programmer's laziness at all.
    ' Honest.
    ' (Well, OK - it's because we need to record line numbers for the logs, which we
    ' didn't do when we were checking the contents of the file in the code above.)
        
    SEEK fileHnd, 1               ' Set the file position
    currentLine = 0               ' Guess what the current line number is?
    DO
        IF readCycleReset = 0 THEN          ' Did we set a flag to reset reading last time around?
            position = SEEK(fileHnd)        ' Set the position marker
            ERRCLEAR                        ' Clear any errors 
            LINE INPUT# fileHnd, bufferIn   ' Read the next line of the file
            INCR currentLine                ' Increment the line number counter
        ELSE
            bufferIn = "BEGIN:VCARD"        ' If the read cycle was reset, then the current line should
                                            ' be "BEGIN:VCARD" anyway...
                                            
            readCycleReset = 0              ' We've reset the read cycle (bypassing the counter
                                            ' increments etc.) so we can mark this flag as unset
        END IF
        
        IF utrim$(bufferIn) = "BEGIN:VCARD" THEN
            
            ' Reset all the temporary variables we use...
            bufferOut = ""
            vCardFileOut = ""
            vCardName = ""
            vCardOrg = ""
            skipvCard = 0
            
            ' We've found a vCard - tell the user!
            STDOUT "   Found vCard at line " & stripNum(currentLine)
            
            ' Who's the vCard for?
            STDOUT "    vCard is for ";
            
            ' Increment the vCard total counter
            INCR cardsFound
            
            ' Add the current text to the output buffer, along with a CR/LF pair...
            bufferOut = bufferOut & bufferIn & $CRLF
            
            ' Until we hit the end of the vCard, we just keep going -
            ' Adding the text to the output buffer
            
            DO UNTIL utrim$(bufferIn) = "END:VCARD"
                
                LINE INPUT# fileHnd, bufferIn                ' Read the next line
                INCR currentLine                             ' Increment the line counter
                bufferOut = bufferOut & bufferIn & $CRLF     ' Add the line to the output buffer
                                 
                ' Is the line a full name or organisation?
                ' If so, we can use it to name the vCard...
                ' Also check for any premature vCard beginnings...
                IF LEFT$(utrim$(bufferIn), 3) = "FN:" THEN 
                    tmpStr = TRIM$(bufferIn)
                    vCardName = RIGHT$(tmpStr, LEN(tmpStr) -3)
                ELSEIF LEFT$(utrim$(bufferIn), 4) = "ORG:" THEN
                    tmpStr = TRIM$(bufferIn)
                    vCardOrg = RIGHT$(tmpStr, LEN(tmpStr) - 4)
                ELSEIF utrim$(bufferIn) = "BEGIN:VCARD" THEN
                    STDOUT "    ERROR! New vCard started before previous one ended!"
                    STDOUT "    Skipping incorrectly terminated entry..."
                    INCR cardsUnfinished                     ' Increment the cardsUnfinished counter
                    INCR cardsSkipped                        ' Increment the cardsSkipped counter
                    readCycleReset = 1                       ' Reset the read cycle to deal with this new card
                    EXIT LOOP
                END IF
                
                ' Did we read past the end of the file?
                IF ERR = 62 THEN
                    vCardEOF = 1              ' Set the premature end of file flag
                    readCycleReset = 1        ' Reset the read cycle, to skip the next loop
                    EXIT LOOP
                END IF
            LOOP
            
            ' If the read cycle has been reset, we can skip this section
            IF readCycleReset = 0 THEN
                
                ' If there's no vCard name, create one -
                '   Use the full name first, then the organisation 
                ' Skip any cards which cannot be named...
                IF vCardName <> "" THEN
                    STDOUT vCardName
                    vCardFileOut = vCardName
                ELSEIF vCardOrg <> "" THEN
                    STDOUT vCardOrg
                    vCardFileOut = vCardOrg
                ELSE
                    STDOUT "unknown entity (No name or organisation!) - skipping!"
                    skipvCard = 1                                    ' Set the skipvCard flag
                    INCR cardsUnusable                               ' Increment the unusable cards flag
                END IF
                
                ' If the vCard has a name, we don't skip it. Otherwise, we skip it.
                IF skipvCard = 0 THEN
                    
                    tmpStr = vCardFileOut
                    
                    ' Get a usable, legal filename - and append the .vcf extension to it
                    vCardFileOut = FormatFileName(tmpStr) & ".vcf"
                    
                    ' Tell the user what we're doing...
                    STDOUT "     Outputting to file: " & CHR$(34) & vCardFileOut & CHR$(34)
                    STDOUT "     Output vCard Length is: " & stripNum(LEN(bufferOut))
                    
                    ' Now write to the file - skip if the file exists, or any unusual errors occur...
                    IF NoFile(vCardFileOut) THEN
                        ERRCLEAR
                        outputFileHnd = FREEFILE
                        OPEN vCardFileOut FOR OUTPUT _               ' Open a file for appending text to
                                          ACCESS WRITE _             ' For reading and writing
                                          LOCK READ WRITE _          ' Nothing else can access it
                                          AS outputfileHnd
                        PRINT# outputFileHnd, bufferOut;
                        IF ERR > 0 THEN
                            STDOUT "!!An error occurred whilst writing to " & vCardFileOut &", the error code was " & stripNum(ERR)
                            INCR cardsWriteError
                        ELSE
                            INCR cardsWritten
                        END IF
                        CLOSE outputFileHnd
                    ELSE
                        STDOUT "     File already exists, or name is invalid - skipping"
                        INCR cardsExist
                        INCR cardsSkipped
                    END IF
                ELSE
                    INCR cardsSkipped
                END IF
            END IF
        END IF
    LOOP UNTIL ERR = 62 ' ERR 62 - reading past end of file

    ' Close the vCard file - we're finished with it...
    CLOSE fileHnd

    ' If the vCard file ended prematurely, inform the user...
    ' And increment the counters appropriately!
    IF vCardEOF = 1 THEN
        STDOUT "     Final vCard did not end correctly - file is truncated in some way!"
        INCR cardsSkipped
        INCR cardsUnfinished
    END IF

    ' Print some quick and dirty stats for the user on the numbers of vCards we handled. 
    STDOUT ""
    STDOUT "--------"
    STDOUT "Finished processing " & CHR$(34) & cmd & CHR$(34)
    STDOUT "Statistics:"
    STDOUT "     vCards found:   " & stripNum(cardsFound)
    STDOUT "     vCards skipped: " & stripNum(cardsSkipped)
    STDOUT "     vCards written: " & stripNum(cardsWritten)
    IF cardsUnfinished > 0 THEN
        IF cardsUnfinished = 1 THEN
            STDOUT "     " & stripNum(cardsUnfinished) & " vCard was incorrectly terminated (truncated)"
        ELSE
            STDOUT "     " & stripNum(cardsUnfinished) & " vCards were incorrectly terminated (truncated)"
        END IF
    END IF
    IF cardsUnusable > 0 THEN
        IF cardsUnusable = 1 THEN
            STDOUT "     " & stripNum(cardsUnusable) & " vCard was skipped as it was unusable (No Name or Organisation entry)"
        ELSE
            STDOUT "     " & stripNum(cardsUnusable) & " vCards were skipped as they were unusable (No Name or Organisation entry)"
        END IF
    END IF
    IF cardsExist > 0 THEN
        IF cardsExist = 1 THEN
            STDOUT "     " & stripNum(cardsExist) & " vCard was skipped as it already existed in this folder"
        ELSE
            STDOUT "     " & stripNum(cardsExist) & " vCards were skipped as they already existed in this folder"
        END IF
    END IF
    IF cardsWriteError > 0 THEN
        IF cardsWriteError = 1 THEN
            STDOUT "     " & stripNum(cardsWriteError) & " vCard experienced an error of some sort on writing"
        ELSE
            STDOUT "     " & stripNum(cardsWriteError) & " vCards experienced an error of some sort on writing"
        END IF
    END IF
    
    ' Print one blank line, to leave some space between us and the command prompt
    STDOUT ""

END FUNCTION


FUNCTION FormatFileName (unformatted AS STRING) AS STRING

    ' This function takes a string and:
    '   * removes any illegal characters from it
    '   * removes any spaces at the start or end
    '   * collapses multiple spaces within the striong to single spaces 
    ' The returned string will be suitable for use as a filename
    
    ' Define the variables we will need
    LOCAL workingStr     AS STRING
    LOCAL tempStr        AS STRING
    LOCAL disallowed     AS STRING
    
    
    ' Set up the working string we will use
    workingStr = unformatted
    
    ' Replace the characters unsuitable for filenames with spaces
    '   disallowed contains what we will replace 
    disallowed = "\/:*?" & CHR$(34) & "<>|"
    REPLACE ANY disallowed WITH "         " IN workingStr

    ' trim the working string of any unnecessary spaces at its start or end
    tempStr = TRIM$(workingStr)
    workingStr = tempStr

    ' Remove any trailing full stops
    DO UNTIL RIGHT$(workingStr, 1) <> "."
        workingStr = LEFT$(workingStr, (LEN(workingStr) -1))
    LOOP

    ' Remove duplicate spaces within the string
    DO UNTIL INSTR(workingStr, "  ") = 0
        REPLACE "  " WITH " " IN workingStr
    LOOP
    
    ' Return the filename we have created...
    FormatFileName = workingStr
    
END FUNCTION

SUB DisplayCopyright ()
    
    ' This subroutine prints the copyright/version information for the program
    
    STDOUT
    STDOUT $ProgramName & " version " & stripNum(%MajorVersion) & "." _
           & stripNum(%MinorVersion) & "." _
           & stripNum(%RevisionNumber) & " (Build " _
           & stripNum(%BuildNumber) & ")." _
           & " Written by " & $ProgramAuthor & "."
    STDOUT " (Written at the request of George R. Watson, who needed it.)"
    STDOUT $ProgramName & " is free software. Copyright " _
           & $ProgramAuthor & " (C) " _
           & $ProgramDate & "."
    STDOUT "For details of the license this program is distributed under,"
    STDOUT "run this program with the command line option " _
           & CHR$(34) & "license" & CHR$(34) & "."
    STDOUT ""
    
    #IF %DevelopmentState = 1
        STDOUT "   !!!This program is an alpha test release - use with caution!!!"
        STDOUT ""
    #ELSEIF %DevelopmentState = 2
        STDOUT "   !This program is a beta test release - use with caution!"
        STDOUT ""
    #ENDIF
    
END SUB

SUB DisplayHelp ()
    
    ' This subroutine prints the help for the program
    
    ' We still need the copyright statement, so let's call it...
    CALL DisplayCopyright ()
    
    STDOUT "   " & $ProgramName & " splits one large combined vCard - as exported by" _
           & " any"
    STDOUT "   application, such as the Palm Desktop - into individual vCard files."
    STDOUT ""
    STDOUT "   The seperate vCard files will be named after either the person's name,"
    STDOUT "   or the organisation name."
    STDOUT "   Any vCards not having a name or organisation name will be not be"
    STDOUT "   extracted - they will be skipped."
    STDOUT ""
    STDOUT "   " & $ProgramName & " does not EVER overwrite any files. If a vCard's " _
           & "filename"
    STDOUT "   already exists, then that vCard will be skipped."
    STDOUT "   A full report of the number of vCards found, written and skipped is"
    STDOUT "   produced when the program stops running."
    STDOUT ""
    STDOUT "   Note that to prevent illegal filenames, the following characters"
    STDOUT "   will be stripped when filenames are generated from fields:"
    STDOUT "      \/:*?" & CHR$(34) & "<>|"
    STDOUT "   and replaced with spaces. Furthermore, duplicate spaces will be"
    STDOUT "   stripped down to one space, and trailing full stops removed. These two"
    STDOUT "   actions are done to create " & CHR$(34) & "prettier" & CHR$(34) & " filenames."
    STDOUT "   Although I would advise you check your contacts/address book for"
    STDOUT "   duplicates and any badly formatted records before exporting for use with"
    STDOUT "   this program, those measures should result in usable filenames"
    STDOUT "   nonetheless."
    STDOUT ""
    STDOUT "   All output from this program is to STDOUT. You can redirect it if you"
    STDOUT "   wish, to produce a form of log. If you receieve unexpected results,"
    STDOUT "   copy the original VCF file you used to somewhere safe and then delete"
    STDOUT "   all the VCF files."
    STDOUT "   Then recopy the original file back and run vCardSplit again, redirecting"
    STDOUT "   the output to a file for logging purposes."
    STDOUT ""
    STDOUT "   Usage:"
    STDOUT "      At the command prompt, type"
    STDOUT "      " & $ProgramName & " <filename>"
    STDOUT "      where <filename> is the .VCF file you wish to split."
    STDOUT ""
    
END SUB

SUB DisplayLicense ()

    ' This subroutine prints the basics of the GNY Public License, under
    ' which this work is released.
    
    ' We still need the copyright to be displayed - call that function first
    CALL DisplayCopyright ()
    
    STDOUT "This program is free software; you can redistribute it and/or"
    STDOUT "modify it under the terms of the GNU General Public License"
    STDOUT "as published by the Free Software Foundation; either version 2"
    STDOUT "of the License, or (at your option) any later version."
    STDOUT ""
    STDOUT "This program is distributed in the hope that it will be useful,"
    STDOUT "but WITHOUT ANY WARRANTY; without even the implied warranty of"
    STDOUT "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the"
    STDOUT "GNU General Public License for more details."
    STDOUT ""
    STDOUT "You should have received a copy of the GNU General Public License"
    STDOUT "along with this program (included in a file called license.txt"
    STDOUT "with all distributions); if not, write to the Free Software"
    STDOUT "Foundation, Inc., 59 Temple Place - Suite 330, Boston,"
    STDOUT "MA  02111-1307, USA."
    STDOUT ""
    
END SUB
